#! /usr/bin/python
# -*- encoding: utf-8 -*-

import torch
import torch.nn.functional as F

def accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k"""
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0, keepdim=True)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res

class PreEmphasis(torch.nn.Module):

    def __init__(self, coef: float = 0.97):
        super().__init__()
        self.coef = coef
        # make kernel
        # In pytorch, the convolution operation uses cross-correlation. So, filter is flipped.
        self.register_buffer(
            'flipped_filter', torch.FloatTensor([-self.coef, 1.]).unsqueeze(0).unsqueeze(0)
        )

    def forward(self, input: torch.tensor) -> torch.tensor:
        assert len(input.size()) == 2, 'The number of dimensions of input tensor must be 2!'
        # reflect padding to match lengths of in/out
        input = input.unsqueeze(1)
        input = F.pad(input, (1, 0), 'reflect')
        return F.conv1d(input, self.flipped_filter).squeeze(1)


#!/usr/bin/python
# -*- coding: utf-8 -*-

import torch
import torch.nn as nn
import torch.nn.functional as F

# from .DatasetLoader import test_dataset_loader
from .deepInfoMaxLoss import DeepInfoMaxLoss
from .Spk_Encoder import MainModel
from .aamsoftmax import LossFunction

class WrappedModel(nn.Module):

    def __init__(self, model):
        super(WrappedModel, self).__init__()
        self.module = model

    def forward(self, x, label=None, l2_reg_dict=None):
        return self.module(x, label)
    

class SpeakerNet(nn.Module):
    """
    SpeakerNet class for speaker recognition neural network.

    forward:
    Forward pass of the SpeakerNet model.

    Args:
        data: Input data for the model.
        label: Labels for the input data (default is None).
        l2_reg_dict: Dictionary for L2 regularization (default is None).

    Returns:
        Total loss, precision, and classification loss.
    """


    def __init__(self, model, optimizer, trainfunc, nPerSpeaker, device, **kwargs):
        super(SpeakerNet, self).__init__();

        self.__S__ = MainModel(device,**kwargs);

        self.__L__ = LossFunction(**kwargs);
        self.__DIM_L__ = DeepInfoMaxLoss(alpha=kwargs['alpha'], beta=kwargs['beta'], gamma=kwargs['gamma'])

        self.nPerSpeaker = nPerSpeaker
        self.weight_finetuning_reg = kwargs['weight_finetuning_reg']
        self.device = device

    def forward(self, data, label=None, l2_reg_dict=None):
        if label is None:
            data_reshape = data[0].to(self.device)
            outp, M = self.__S__.forward([data_reshape, data[1]])
            return outp
        else:
            data_reshape = data[0].reshape(-1, data[0].size()[-1]).to(self.device)
            outp, M = self.__S__.forward([data_reshape, data[1]])

            M = M.transpose(0, 1).transpose(1, 2)

            M_prime = torch.cat((M[1:], M[0].unsqueeze(0)), dim=0)

            nloss, prec1 = self.__L__.forward(outp, label)

            if l2_reg_dict is not None:
                Learned_dict = l2_reg_dict
                l2_reg = 0
                for name, param in self.__S__.model.named_parameters():
                    if name in Learned_dict:
                        l2_reg = l2_reg + torch.norm(param - Learned_dict[name].cuda(), 2)
                tloss = nloss / nloss.detach() + self.weight_finetuning_reg * l2_reg / (l2_reg.detach() + 1e-5)
            else:
                tloss = nloss
                # print("Without L2 Reg")

            dim_loss = self.__DIM_L__(outp, M, M_prime)

            # add loss
            t_loss = tloss + dim_loss

            print(f"classification Loss: {tloss}, skloss {nloss} , DIM Loss: {dim_loss} total loss {t_loss} \n")
            return t_loss, prec1, nloss



import os
from huggingface_hub import hf_hub_download
import torch

def get_default_param():
    """
    Get the default parameters for the model.

    Returns:
        Dictionary containing default parameters for the model.
    """

    args = {'config': None, 
            'max_frames': 300, 
            'eval_frames': 400, 
            'batch_size': 32, 
            'max_seg_per_spk': 500, 
            'nDataLoaderThread': 10, 
            'augment': False, 
            'seed': 20211202, 
            'test_interval': 1, 
            'max_epoch': 15, 
            'trainfunc': 'aamsoftmax', 
            'optimizer': 'adamw', 
            'scheduler': 'steplr', 
            'lr': 0.001, 
            'lr_decay': 0.95, 
            'pretrained_model_path': '/home/$USER/.cache/huggingface/hub/models--thejan-fonseka--DeepSpeakerVerifier/snapshots/*/WavLM-Base+.pt', 
            'weight_finetuning_reg': 0.01, 
            'LLRD_factor': 1.0, 
            'LR_Transformer': 2e-05, 
            'LR_MHFA': 0.005, 
            'hard_prob': 0.5, 
            'hard_rank': 10, 
            'margin': 0.2, 
            'scale': 30, 
            'nPerSpeaker': 1, 
            'nClasses': 1211, 
            'alpha': 0.01, 
            'beta': 0.05, 
            'gamma': 0.005, 
            'dcf_p_target': 0.05, 
            'dcf_c_miss': 1, 
            'dcf_c_fa': 1, 
            'initial_model': '', 
            'save_path': 'exps/exp1', 
            'train_list': '/mnt/proj3/open-24-5/pengjy_new/WavLM/vox_list/train_list.txt', 
            'test_list': '/mnt/proj3/open-24-5/pengjy_new/WavLM/vox_list/veri_test.txt', 
            'train_list_percentage': 1.0, 
            'train_path': '/mnt/proj3/open-24-5/plchot/data_wav/16kHz/voxceleb2/dev/aac/', 
            'test_path': '/mnt/proj3/open-24-5/plchot/data_wav/16kHz/voxceleb_1.1/', 
            'musan_path': '/mnt/proj3/open-24-5/pengjy_new/musan_split/', 
            'rir_path': '/mnt/proj3/open-24-5/plchot/data_augment/16kHz/simulated_rirs/', 
            'n_mels': 80, 
            'log_input': False, 
            'model': '', 
            'encoder_type': 
            'SAP', 'nOut': 256, 
            'eval': False, 
            'save_model_pt': False, 
            'port': '7888', 
            'distributed': False, 
            'mixedprec': False, 
            'attention_heads': 8, 
            'gpu': 0}

    device = "cuda" if torch.cuda.is_available() else "cpu"
    args['device'] = device

    if not os.path.exists(args["pretrained_model_path"]):
        model_file = hf_hub_download(repo_id="thejan-fonseka/DeepSpeakerVerifier", filename="WavLM-Base+.pt")
        args["pretrained_model_path"] = model_file

    return args

if __name__ == "__main__":
    get_default_param()


import torch
import torch.nn as nn
import torch.nn.functional as F
from .WavLM import *

import torch
import torch.nn as nn

class MHFA(nn.Module):
    """
    MHFA class for Multi-Head Feature Aggregation.

    forward:
    Perform the forward pass of the MHFA model.

    Args:
        x: Input tensor.

    Returns:
        Output tensor after feature aggregation.
    """

    def __init__(self, head_nb=8, inputs_dim=768, compression_dim=128, outputs_dim=256):
        super(MHFA, self).__init__()

        self.weights_k = nn.Parameter(data=torch.ones(13), requires_grad=True)
        self.weights_v = nn.Parameter(data=torch.ones(13), requires_grad=True)

        self.head_nb = head_nb
        self.ins_dim = inputs_dim
        self.cmp_dim = compression_dim
        self.ous_dim = outputs_dim

        self.cmp_linear_k = nn.Linear(self.ins_dim, self.cmp_dim)
        self.cmp_linear_v = nn.Linear(self.ins_dim, self.cmp_dim)

        self.att_head = nn.Linear(self.cmp_dim, self.head_nb)
        self.pooling_fc = nn.Linear(self.head_nb * self.cmp_dim, self.ous_dim)

    def forward(self, x):

        k = torch.sum(x.mul(nn.functional.softmax(self.weights_k, dim=-1)), dim=-1).transpose(1, 2)
        v = torch.sum(x.mul(nn.functional.softmax(self.weights_v, dim=-1)), dim=-1).transpose(1, 2)

        k = self.cmp_linear_k(k)
        v = self.cmp_linear_v(v)

        att_k = self.att_head(k)
        v = v.unsqueeze(-2)

        pooling_outs = torch.sum(v.mul(nn.functional.softmax(att_k, dim=1).unsqueeze(-1)), dim=1)

        b, h, f = pooling_outs.shape
        pooling_outs = pooling_outs.reshape(b, -1)

        outs = self.pooling_fc(pooling_outs)

        return outs


class spk_extractor(nn.Module):
    """
    spk_extractor class for speaker feature extraction.

    forward:
    Perform the forward pass of the spk_extractor model.

    Args:
        wav_and_flag: Input tensor containing audio data and flag.

    Returns:
        Output tensor and last layer representation.
    """

    def __init__(self,device,**kwargs):
        super(spk_extractor, self).__init__()
        # print("Pre-trained Model: {}".format(kwargs['pretrained_model_path']))
        checkpoint = torch.load(kwargs['pretrained_model_path'], map_location=torch.device(device))
        cfg = WavLMConfig(checkpoint['cfg'])
        # cfg = WavLMConfig()
        self.model = WavLM(cfg)
        self.loadParameters(checkpoint['model'])
        head_nb = kwargs['attention_heads']
        self.backend = MHFA(head_nb=head_nb)
        # print(f"Number of Heads : {head_nb}")


    def forward(self,wav_and_flag):
        
        x = wav_and_flag[0]

        cnn_outs, layer_results, last_layer =  self.model.extract_features(x, output_layer=13)
        layer_reps = [x.transpose(0, 1) for x, _ in layer_results]
       
        x = torch.stack(layer_reps).transpose(0,-1).transpose(0,1)
        out = self.backend(x)
        return out, last_layer

    def loadParameters(self, param):

        self_state = self.model.state_dict();
        loaded_state = param

        for name, param in loaded_state.items():
            origname = name;
            
            if name not in self_state:
                # print("%s is not in the model."%origname);
                continue;
            if self_state[name].size() != loaded_state[origname].size():
                print("Wrong parameter length: %s, model: %s, loaded: %s"%(origname, self_state[name].size(), loaded_state[origname].size()));
                continue;
            self_state[name].copy_(param);

def MainModel(device,**kwargs):
    model = spk_extractor(device, **kwargs)
    return model


import torch
import torch.nn.functional as F
from torch import nn as nn


class GlobalDiscriminator(nn.Module):
    def __init__(self, M_channels=150, interm_size=512):
        super().__init__()

        self.c0 = torch.nn.Conv1d(768, 256, kernel_size=3)
        self.c1 = torch.nn.Conv1d(256, 64, kernel_size=3)
        # see appendix 1A of https://arxiv.org/pdf/1808.06670.pdf
        # input of self.l0 is the concatenate of E and flattened output of self.c1 (C)
        in_feature = 64 * (150-2*2) + 256
        self.l0 = torch.nn.Linear(in_feature, interm_size)
        self.l1 = torch.nn.Linear(interm_size, interm_size)
        self.l2 = torch.nn.Linear(interm_size, 1)

    def forward(self, E, M):

        C = F.relu(self.c0(M))
        C = self.c1(C)
        C = C.view(E.shape[0], -1)
        out = torch.cat((E, C), dim=1)
        out = F.relu(self.l0(out))
        out = F.relu(self.l1(out))
        out = self.l2(out)
        # see appendix 1A of https://arxiv.org/pdf/1808.06670.pdf
        # output of Table 5
        return out

class LocalDiscriminator(nn.Module):

    def __init__(self, interm_channels=512):
        super().__init__()
        
        self.c0 = torch.nn.Conv1d(256+768, interm_channels, kernel_size=1)
        self.c1 = torch.nn.Conv1d(interm_channels, interm_channels, kernel_size=1)
        self.c2 = torch.nn.Conv1d(interm_channels, 1, kernel_size=1)

    def forward(self, x):
        score = F.relu(self.c0(x))
        score = F.relu(self.c1(score))
        score = self.c2(score)
        return score
    
class PriorDiscriminator(nn.Module):
    def __init__(self):
        super().__init__()
        self.l0 = nn.Linear(256, 1000)
        self.l1 = nn.Linear(1000, 200)
        self.l2 = nn.Linear(200, 1)

    def forward(self, x):
        h = F.relu(self.l0(x))
        h = F.relu(self.l1(h))
        h = torch.sigmoid(self.l2(h))
        h = torch.clamp(h, min=9e-6)
        return h



import torch
import torch.nn as nn
import torch.nn.functional as F

from .mine import GlobalDiscriminator, LocalDiscriminator, PriorDiscriminator


class DeepInfoMaxLoss(nn.Module):
    def __init__(self, alpha, beta, gamma):
        super().__init__()
        self.global_d = GlobalDiscriminator()
        self.local_d = LocalDiscriminator()
        self.prior_d = PriorDiscriminator()
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma

    def forward(self, y, M, M_prime):

        # see appendix 1A of https://arxiv.org/pdf/1808.06670.pdf
        y_exp_b = y.unsqueeze(-1)
        y_exp = y_exp_b.expand(-1, -1, 150)

        y_M = torch.cat((M, y_exp), dim=1)
        y_M_prime = torch.cat((M_prime, y_exp), dim=1)
        
        Ej = -F.softplus(-self.local_d(y_M)).mean()
        Em = F.softplus(self.local_d(y_M_prime)).mean()
        LOCAL = (Em - Ej) * self.beta

        Ej_global = -F.softplus(-self.global_d(y, M)).mean()
        Em_global = F.softplus(self.global_d(y, M_prime)).mean()
        GLOBAL = (Em_global - Ej_global) * self.alpha

        prior = torch.rand_like(y)

        term_a = torch.log(self.prior_d(prior)).mean()
        term_b = torch.log(1.0 - self.prior_d(y)).mean()
        PRIOR = - (term_a + term_b) * self.gamma
        
        return LOCAL + GLOBAL + PRIOR



# --------------------------------------------------------
# WavLM: Large-Scale Self-Supervised  Pre-training  for Full Stack Speech Processing (https://arxiv.org/abs/2110.13900.pdf)
# Github source: https://github.com/microsoft/unilm/tree/master/wavlm
# Copyright (c) 2021 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Based on fairseq code bases
# https://github.com/pytorch/fairseq
# --------------------------------------------------------

import logging
import math
from typing import List, Optional, Tuple

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import LayerNorm

from .modules import (
    Fp32GroupNorm,
    Fp32LayerNorm,
    MultiheadAttention,
    SamePad,
    init_bert_params,
    get_activation_fn,
    TransposeLast,
    GLU_Linear,
)

logger = logging.getLogger(__name__)


def compute_mask_indices(
    shape: Tuple[int, int],
    padding_mask: Optional[torch.Tensor],
    mask_prob: float,
    mask_length: int,
    mask_type: str = "static",
    mask_other: float = 0.0,
    min_masks: int = 0,
    no_overlap: bool = False,
    min_space: int = 0,
) -> np.ndarray:
    """
    Computes random mask spans for a given shape

    Args:
        shape: the the shape for which to compute masks.
            should be of size 2 where first element is batch size and 2nd is timesteps
        padding_mask: optional padding mask of the same size as shape, which will prevent masking padded elements
        mask_prob: probability for each token to be chosen as start of the span to be masked. this will be multiplied by
            number of timesteps divided by length of mask span to mask approximately this percentage of all elements.
            however due to overlaps, the actual number will be smaller (unless no_overlap is True)
        mask_type: how to compute mask lengths
            static = fixed size
            uniform = sample from uniform distribution [mask_other, mask_length*2]
            normal = sample from normal distribution with mean mask_length and stdev mask_other. mask is min 1 element
            poisson = sample from possion distribution with lambda = mask length
        min_masks: minimum number of masked spans
        no_overlap: if false, will switch to an alternative recursive algorithm that prevents spans from overlapping
        min_space: only used if no_overlap is True, this is how many elements to keep unmasked between spans
    """

    bsz, all_sz = shape
    mask = np.full((bsz, all_sz), False)

    all_num_mask = int(
        # add a random number for probabilistic rounding
        mask_prob * all_sz / float(mask_length)
        + np.random.rand()
    )

    all_num_mask = max(min_masks, all_num_mask)

    mask_idcs = []
    for i in range(bsz):
        if padding_mask is not None:
            sz = all_sz - padding_mask[i].long().sum().item()
            num_mask = int(
                # add a random number for probabilistic rounding
                mask_prob * sz / float(mask_length)
                + np.random.rand()
            )
            num_mask = max(min_masks, num_mask)
        else:
            sz = all_sz
            num_mask = all_num_mask

        if mask_type == "static":
            lengths = np.full(num_mask, mask_length)
        elif mask_type == "uniform":
            lengths = np.random.randint(mask_other, mask_length * 2 + 1, size=num_mask)
        elif mask_type == "normal":
            lengths = np.random.normal(mask_length, mask_other, size=num_mask)
            lengths = [max(1, int(round(x))) for x in lengths]
        elif mask_type == "poisson":
            lengths = np.random.poisson(mask_length, size=num_mask)
            lengths = [int(round(x)) for x in lengths]
        else:
            raise Exception("unknown mask selection " + mask_type)

        if sum(lengths) == 0:
            lengths[0] = min(mask_length, sz - 1)

        if no_overlap:
            mask_idc = []

            def arrange(s, e, length, keep_length):
                span_start = np.random.randint(s, e - length)
                mask_idc.extend(span_start + i for i in range(length))

                new_parts = []
                if span_start - s - min_space >= keep_length:
                    new_parts.append((s, span_start - min_space + 1))
                if e - span_start - keep_length - min_space > keep_length:
                    new_parts.append((span_start + length + min_space, e))
                return new_parts

            parts = [(0, sz)]
            min_length = min(lengths)
            for length in sorted(lengths, reverse=True):
                lens = np.fromiter(
                    (e - s if e - s >= length + min_space else 0 for s, e in parts),
                    np.int,
                )
                l_sum = np.sum(lens)
                if l_sum == 0:
                    break
                probs = lens / np.sum(lens)
                c = np.random.choice(len(parts), p=probs)
                s, e = parts.pop(c)
                parts.extend(arrange(s, e, length, min_length))
            mask_idc = np.asarray(mask_idc)
        else:
            min_len = min(lengths)
            if sz - min_len <= num_mask:
                min_len = sz - num_mask - 1

            mask_idc = np.random.choice(sz - min_len, num_mask, replace=False)

            mask_idc = np.asarray(
                [
                    mask_idc[j] + offset
                    for j in range(len(mask_idc))
                    for offset in range(lengths[j])
                ]
            )

        mask_idcs.append(np.unique(mask_idc[mask_idc < sz]))

    min_len = min([len(m) for m in mask_idcs])
    for i, mask_idc in enumerate(mask_idcs):
        if len(mask_idc) > min_len:
            mask_idc = np.random.choice(mask_idc, min_len, replace=False)
        mask[i, mask_idc] = True

    return mask


class WavLMConfig:
    def __init__(self, cfg=None):
        self.extractor_mode: str = "default"     # mode for feature extractor. default has a single group norm with d groups in the first conv block, whereas layer_norm has layer norms in every block (meant to use with normalize=True)
        self.encoder_layers: int = 12     # num encoder layers in the transformer

        self.encoder_embed_dim: int = 768     # encoder embedding dimension
        self.encoder_ffn_embed_dim: int = 3072     # encoder embedding dimension for FFN
        self.encoder_attention_heads: int = 12     # num encoder attention heads
        self.activation_fn: str = "gelu"     # activation function to use

        self.layer_norm_first: bool = False     # apply layernorm first in the transformer
        self.conv_feature_layers: str = "[(512,10,5)] + [(512,3,2)] * 4 + [(512,2,2)] * 2"     # string describing convolutional feature extraction layers in form of a python list that contains [(dim, kernel_size, stride), ...]
        self.conv_bias: bool = False     # include bias in conv encoder
        self.feature_grad_mult: float = 1.0     # multiply feature extractor var grads by this

        self.normalize: bool = False  # normalize input to have 0 mean and unit variance during training

        # dropouts
        self.dropout: float = 0.1     # dropout probability for the transformer
        self.attention_dropout: float = 0.1     # dropout probability for attention weights
        self.activation_dropout: float = 0.0     # dropout probability after activation in FFN
        self.encoder_layerdrop: float = 0.0     # probability of dropping a tarnsformer layer
        self.dropout_input: float = 0.0     # dropout to apply to the input (after feat extr)
        self.dropout_features: float = 0.0     # dropout to apply to the features (after feat extr)

        # masking
        self.mask_length: int = 10     # mask length
        self.mask_prob: float = 0.65     # probability of replacing a token with mask
        self.mask_selection: str = "static"     # how to choose mask length
        self.mask_other: float = 0     # secondary mask argument (used for more complex distributions), see help in compute_mask_indicesh
        self.no_mask_overlap: bool = False     # whether to allow masks to overlap
        self.mask_min_space: int = 1     # min space between spans (if no overlap is enabled)

        # channel masking
        self.mask_channel_length: int = 10     # length of the mask for features (channels)
        self.mask_channel_prob: float = 0.0     # probability of replacing a feature with 0
        self.mask_channel_selection: str = "static"     # how to choose mask length for channel masking
        self.mask_channel_other: float = 0     # secondary mask argument (used for more complex distributions), see help in compute_mask_indices
        self.no_mask_channel_overlap: bool = False     # whether to allow channel masks to overlap
        self.mask_channel_min_space: int = 1     # min space between spans (if no overlap is enabled)

        # positional embeddings
        self.conv_pos: int = 128     # number of filters for convolutional positional embeddings
        self.conv_pos_groups: int = 16     # number of groups for convolutional positional embedding

        # relative position embedding
        self.relative_position_embedding: bool = False     # apply relative position embedding
        self.num_buckets: int = 320     # number of buckets for relative position embedding
        self.max_distance: int = 1280     # maximum distance for relative position embedding
        self.gru_rel_pos: bool = False     # apply gated relative position embedding

        if cfg is not None:
            self.update(cfg)

    def update(self, cfg: dict):
        self.__dict__.update(cfg)


class WavLM(nn.Module):
    def __init__(
        self,
        cfg: WavLMConfig,
    ) -> None:
        super().__init__()
        logger.info(f"WavLM Config: {cfg.__dict__}")

        self.cfg = cfg
        feature_enc_layers = eval(cfg.conv_feature_layers)
        self.embed = feature_enc_layers[-1][0]
        # print(f"\n\n\n\n\n\ {feature_enc_layers=} \n\n\n\n\n\n")

        self.feature_extractor = ConvFeatureExtractionModel(
            conv_layers=feature_enc_layers,
            dropout=0.0,
            mode=cfg.extractor_mode,
            conv_bias=cfg.conv_bias,
        )

        # print(f"self.feature extract \n\n\n\n {self.feature_extractor}\n\n\n\n\n\n")

        self.post_extract_proj = (
            nn.Linear(self.embed, cfg.encoder_embed_dim)
            if self.embed != cfg.encoder_embed_dim
            else None
        )

        self.mask_prob = cfg.mask_prob
        self.mask_selection = cfg.mask_selection
        self.mask_other = cfg.mask_other
        self.mask_length = cfg.mask_length
        self.no_mask_overlap = cfg.no_mask_overlap
        self.mask_min_space = cfg.mask_min_space

        self.mask_channel_prob = cfg.mask_channel_prob
        self.mask_channel_selection = cfg.mask_channel_selection
        self.mask_channel_other = cfg.mask_channel_other
        self.mask_channel_length = cfg.mask_channel_length
        self.no_mask_channel_overlap = cfg.no_mask_channel_overlap
        self.mask_channel_min_space = cfg.mask_channel_min_space

        self.dropout_input = nn.Dropout(cfg.dropout_input)
        self.dropout_features = nn.Dropout(cfg.dropout_features)

        self.feature_grad_mult = cfg.feature_grad_mult

        self.mask_emb = nn.Parameter(
            torch.FloatTensor(cfg.encoder_embed_dim).uniform_()
        )

        self.encoder = TransformerEncoder(cfg)
        self.layer_norm = LayerNorm(self.embed)

    def apply_mask(self, x, padding_mask):
        B, T, C = x.shape
        if self.mask_prob > 0:
            mask_indices = compute_mask_indices(
                (B, T),
                padding_mask,
                self.mask_prob,
                self.mask_length,
                self.mask_selection,
                self.mask_other,
                min_masks=2,
                no_overlap=self.no_mask_overlap,
                min_space=self.mask_min_space,
            )
            mask_indices = torch.from_numpy(mask_indices).to(x.device)
            x[mask_indices] = self.mask_emb
        else:
            mask_indices = None

        if self.mask_channel_prob > 0:
            mask_channel_indices = compute_mask_indices(
                (B, C),
                None,
                self.mask_channel_prob,
                self.mask_channel_length,
                self.mask_channel_selection,
                self.mask_channel_other,
                no_overlap=self.no_mask_channel_overlap,
                min_space=self.mask_channel_min_space,
            )
            mask_channel_indices = (
                torch.from_numpy(mask_channel_indices)
                .to(x.device)
                .unsqueeze(1)
                .expand(-1, T, -1)
            )
            x[mask_channel_indices] = 0

        return x, mask_indices

    def forward_padding_mask(
            self, features: torch.Tensor, padding_mask: torch.Tensor,
    ) -> torch.Tensor:
        extra = padding_mask.size(1) % features.size(1)
        if extra > 0:
            padding_mask = padding_mask[:, :-extra]
        padding_mask = padding_mask.view(
            padding_mask.size(0), features.size(1), -1
        )
        padding_mask = padding_mask.all(-1)
        return padding_mask

    def extract_features(
        self,
        source: torch.Tensor,
        padding_mask: Optional[torch.Tensor] = None,
        mask: bool = False,
        ret_conv: bool = False,
        output_layer: Optional[int] = None,
        ret_layer_results: bool = False,
    ):


        with torch.no_grad():
            features = self.feature_extractor(source)
        # print(f"eevefeefef features {features[0].shape=}")
        # print(f"eevefeefef features {features[-1].shape=}")


        cnn_outs = features
        features = features[-1].transpose(1, 2)
        features = self.layer_norm(features)

        if padding_mask is not None:
            padding_mask = self.forward_padding_mask(features, padding_mask)

        if self.post_extract_proj is not None:
            features = self.post_extract_proj(features)

        features = self.dropout_input(features)

        if mask:
            x, mask_indices = self.apply_mask(
                features, padding_mask
            )
        else:
            x = features
        # print(f"vvvvvvvvvvvvvvvvvv === {x.shape=}")
        # feature: (B, T, D), float
        # target: (B, T), long
        # x: (B, T, D), float
        # padding_mask: (B, T), bool
        # mask_indices: (B, T), bool
        # print(f"\n\n\n\n\n\n\n\n {padding_mask=} \n\n\n\n\n\n\n ")
        x, layer_results, last_layer = self.encoder(
            x,
            padding_mask=padding_mask,
            layer=None if output_layer is None else output_layer - 1
        )
        # print(f"vcvbbbbbbbbbbb  {x.shape=}")
        # print(f"last layer wvlm = {last_layer.shape}")
        return cnn_outs, layer_results, last_layer
        # res = {"x": x, "padding_mask": padding_mask, "features": features, "layer_results": layer_results}

        # feature = res["features"] if ret_conv else res["x"]
        # if ret_layer_results:
        #     feature = (feature, res["layer_results"])
        # return feature, res["padding_mask"]


class ConvFeatureExtractionModel(nn.Module):
    def __init__(
            self,
            conv_layers: List[Tuple[int, int, int]],
            dropout: float = 0.0,
            mode: str = "default",
            conv_bias: bool = False,
            conv_type: str = "default"
    ):
        super().__init__()

        assert mode in {"default", "layer_norm"}

        def block(
                n_in,
                n_out,
                k,
                stride,
                is_layer_norm=False,
                is_group_norm=False,
                conv_bias=False,
        ):
            def make_conv():
                conv = nn.Conv1d(n_in, n_out, k, stride=stride, bias=conv_bias)
                nn.init.kaiming_normal_(conv.weight)
                return conv

            assert (
                           is_layer_norm and is_group_norm
                   ) == False, "layer norm and group norm are exclusive"

            if is_layer_norm:
                return nn.Sequential(
                    make_conv(),
                    nn.Dropout(p=dropout),
                    nn.Sequential(
                        TransposeLast(),
                        Fp32LayerNorm(dim, elementwise_affine=True),
                        TransposeLast(),
                    ),
                    nn.GELU(),
                )
            elif is_group_norm:
                return nn.Sequential(
                    make_conv(),
                    nn.Dropout(p=dropout),
                    Fp32GroupNorm(dim, dim, affine=True),
                    nn.GELU(),
                )
            else:
                return nn.Sequential(make_conv(), nn.Dropout(p=dropout), nn.GELU())

        self.conv_type = conv_type
        if self.conv_type == "default":
            in_d = 1
            self.conv_layers = nn.ModuleList()
            for i, cl in enumerate(conv_layers):
                assert len(cl) == 3, "invalid conv definition: " + str(cl)
                (dim, k, stride) = cl

                self.conv_layers.append(
                    block(
                        in_d,
                        dim,
                        k,
                        stride,
                        is_layer_norm=mode == "layer_norm",
                        is_group_norm=mode == "default" and i == 0,
                        conv_bias=conv_bias,
                    )
                )
                in_d = dim
        elif self.conv_type == "conv2d":
            in_d = 1
            self.conv_layers = nn.ModuleList()
            for i, cl in enumerate(conv_layers):
                assert len(cl) == 3
                (dim, k, stride) = cl

                self.conv_layers.append(
                    torch.nn.Conv2d(in_d, dim, k, stride)
                )
                self.conv_layers.append(torch.nn.ReLU())
                in_d = dim
        elif self.conv_type == "custom":
            in_d = 1
            idim = 80
            self.conv_layers = nn.ModuleList()
            for i, cl in enumerate(conv_layers):
                assert len(cl) == 3
                (dim, k, stride) = cl
                self.conv_layers.append(
                    torch.nn.Conv2d(in_d, dim, k, stride, padding=1)
                )
                self.conv_layers.append(
                    torch.nn.LayerNorm([dim, idim])
                )
                self.conv_layers.append(torch.nn.ReLU())
                in_d = dim
                if (i + 1) % 2 == 0:
                    self.conv_layers.append(
                        torch.nn.MaxPool2d(2, stride=2, ceil_mode=True)
                    )
                    idim = int(math.ceil(idim / 2))
        else:
            pass

    def forward(self, x, mask=None):

        # BxT -> BxCxT
        # print(f"rvtrbtrb ==== {x.shape=}")
        x_lst = []
        x = x.unsqueeze(1)
        if self.conv_type == "custom":
            for conv in self.conv_layers:
                if isinstance(conv, nn.LayerNorm):
                    x = x.transpose(1, 2)
                    x = conv(x).transpose(1, 2)
                else:
                    x = conv(x)

            x = x.transpose(2, 3).contiguous()
            # print(f"revtbrtbrbtrtvtrbtbt b {x.size(0)}, c {x.shape}, t {x.size(-1)}")
            x = x.view(x.size(0), -1, x.size(-1))
        else:
            for conv in self.conv_layers:
                x = conv(x)
                x_lst.append(x)
            if self.conv_type == "conv2d":
                b, c, t, f = x.size()
                # print(f"revtbrtbrbtrtvtrbtbt b {b}, c {c}, t {t}, f {f}")
                x = x.transpose(2, 3).contiguous().view(b, c * f, t)
            # print(f"revtbrtbr ============={x.size()}")
        # print(f"frfrfrfrf======= {x_lst[0].shape=} || {len(x_lst)=}")
        # for i in range(len(x_lst)):
        #     print(f"ffffffffffff || {i+1} and {x_lst[i].shape=}")
        return x_lst


class TransformerEncoder(nn.Module):
    def __init__(self, args):
        super().__init__()

        self.dropout = args.dropout
        self.embedding_dim = args.encoder_embed_dim

        self.pos_conv = nn.Conv1d(
            self.embedding_dim,
            self.embedding_dim,
            kernel_size=args.conv_pos,
            padding=args.conv_pos // 2,
            groups=args.conv_pos_groups,
        )
        dropout = 0
        std = math.sqrt((4 * (1.0 - dropout)) / (args.conv_pos * self.embedding_dim))
        nn.init.normal_(self.pos_conv.weight, mean=0, std=std)
        nn.init.constant_(self.pos_conv.bias, 0)

        self.pos_conv = nn.utils.parametrizations.weight_norm(self.pos_conv, name="weight", dim=2)
        self.pos_conv = nn.Sequential(self.pos_conv, SamePad(args.conv_pos), nn.GELU())

        if hasattr(args, "relative_position_embedding"):
            self.relative_position_embedding = args.relative_position_embedding
            self.num_buckets = args.num_buckets
            self.max_distance = args.max_distance
        else:
            self.relative_position_embedding = False
            self.num_buckets = 0
            self.max_distance = 0

        self.layers = nn.ModuleList(
            [
                TransformerSentenceEncoderLayer(
                    embedding_dim=self.embedding_dim,
                    ffn_embedding_dim=args.encoder_ffn_embed_dim,
                    num_attention_heads=args.encoder_attention_heads,
                    dropout=self.dropout,
                    attention_dropout=args.attention_dropout,
                    activation_dropout=args.activation_dropout,
                    activation_fn=args.activation_fn,
                    layer_norm_first=args.layer_norm_first,
                    has_relative_attention_bias=(self.relative_position_embedding and i == 0),
                    num_buckets=self.num_buckets,
                    max_distance=self.max_distance,
                    gru_rel_pos=args.gru_rel_pos,
                )
                for i in range(args.encoder_layers)
            ]
        )

        self.layer_norm_first = args.layer_norm_first
        self.layer_norm = LayerNorm(self.embedding_dim)
        self.layerdrop = args.encoder_layerdrop

        self.apply(init_bert_params)

    def forward(self, x, padding_mask=None, streaming_mask=None, layer=None):
        # print(f"trans encode x = {x}   \n\n\n\n {x.shape}")
        # print(f"trans encode layer = {layer}")

        x, layer_results, last_layer = self.extract_features(x, padding_mask, streaming_mask, layer)
        # print(f"trans encode after x = {x} \n\n\n\n {x.shape}")
        # print(f"trans encode after layer = {layer}")
        if self.layer_norm_first and layer is None:
            x = self.layer_norm(x)

        return x, layer_results, last_layer

    def extract_features(self, x, padding_mask=None, streaming_mask=None, tgt_layer=None):

        if padding_mask is not None:
            x[padding_mask] = 0

        x_conv = self.pos_conv(x.transpose(1, 2))
        x_conv = x_conv.transpose(1, 2)
        x = x + x_conv

        if not self.layer_norm_first:
            x = self.layer_norm(x)

        x = F.dropout(x, p=self.dropout, training=self.training)
        # print(f"xxxxxxxxxxxx {x}   \n\n\n\n {x.shape}")
        # B x T x C -> T x B x C
        # print(f"x  B x T x C = {x.shape}")
        x = x.transpose(0, 1)

        # print(f"xxxxxxxxxxxx trans    \n\n\n\n {x.shape}")

        layer_results = []
        z = None
        if tgt_layer is not None:
            layer_results.append((x, z))
        r = None
        pos_bias = None
        for i, layer in enumerate(self.layers):
            dropout_probability = np.random.random()
            if not self.training or (dropout_probability > self.layerdrop):
                x, z, pos_bias = layer(x, self_attn_padding_mask=padding_mask, need_weights=False,
                                       self_attn_mask=streaming_mask, pos_bias=pos_bias)
            if tgt_layer is not None:
                layer_results.append((x, z))
            if i == tgt_layer:
                r = x
                break

        if r is not None:
            x = r

        # T x B x C -> B x T x C
        x = x.transpose(0, 1)

        return x, layer_results, layer_results[-1][0]


class TransformerSentenceEncoderLayer(nn.Module):
    """
    Implements a Transformer Encoder Layer used in BERT/XLM style pre-trained
    models.
    """

    def __init__(
            self,
            embedding_dim: float = 768,
            ffn_embedding_dim: float = 3072,
            num_attention_heads: float = 8,
            dropout: float = 0.1,
            attention_dropout: float = 0.1,
            activation_dropout: float = 0.1,
            activation_fn: str = "relu",
            layer_norm_first: bool = False,
            has_relative_attention_bias: bool = False,
            num_buckets: int = 0,
            max_distance: int = 0,
            rescale_init: bool = False,
            gru_rel_pos: bool = False,
    ) -> None:

        super().__init__()
        # Initialize parameters
        self.embedding_dim = embedding_dim
        self.dropout = dropout
        self.activation_dropout = activation_dropout

        # Initialize blocks
        self.activation_name = activation_fn
        self.activation_fn = get_activation_fn(activation_fn)
        self.self_attn = MultiheadAttention(
            self.embedding_dim,
            num_attention_heads,
            dropout=attention_dropout,
            self_attention=True,
            has_relative_attention_bias=has_relative_attention_bias,
            num_buckets=num_buckets,
            max_distance=max_distance,
            rescale_init=rescale_init,
            gru_rel_pos=gru_rel_pos,
        )

        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(self.activation_dropout)
        self.dropout3 = nn.Dropout(dropout)

        self.layer_norm_first = layer_norm_first

        # layer norm associated with the self attention layer
        self.self_attn_layer_norm = LayerNorm(self.embedding_dim)

        if self.activation_name == "glu":
            self.fc1 = GLU_Linear(self.embedding_dim, ffn_embedding_dim, "swish")
        else:
            self.fc1 = nn.Linear(self.embedding_dim, ffn_embedding_dim)
        self.fc2 = nn.Linear(ffn_embedding_dim, self.embedding_dim)

        # layer norm associated with the position wise feed-forward NN
        self.final_layer_norm = LayerNorm(self.embedding_dim)

    def forward(
            self,
            x: torch.Tensor,
            self_attn_mask: torch.Tensor = None,
            self_attn_padding_mask: torch.Tensor = None,
            need_weights: bool = False,
            pos_bias=None
    ):
        """
        LayerNorm is applied either before or after the self-attention/ffn
        modules similar to the original Transformer imlementation.
        """
        residual = x

        if self.layer_norm_first:
            x = self.self_attn_layer_norm(x)
            x, attn, pos_bias = self.self_attn(
                query=x,
                key=x,
                value=x,
                key_padding_mask=self_attn_padding_mask,
                need_weights=False,
                attn_mask=self_attn_mask,
                position_bias=pos_bias
            )
            x = self.dropout1(x)
            x = residual + x

            residual = x

            x = self.final_layer_norm(x)
            if self.activation_name == "glu":
                x = self.fc1(x)
            else:
                x = self.activation_fn(self.fc1(x))
            x = self.dropout2(x)
            x = self.fc2(x)
            x = self.dropout3(x)
            x = residual + x
        else:

            x, attn, pos_bias = self.self_attn(
                query=x,
                key=x,
                value=x,
                key_padding_mask=self_attn_padding_mask,
                need_weights=need_weights,
                attn_mask=self_attn_mask,
                position_bias=pos_bias
            )

            x = self.dropout1(x)
            x = residual + x

            x = self.self_attn_layer_norm(x)

            residual = x
            if self.activation_name == "glu":
                x = self.fc1(x)
            else:
                x = self.activation_fn(self.fc1(x))
            x = self.dropout2(x)
            x = self.fc2(x)
            x = self.dropout3(x)
            x = residual + x

            x = self.final_layer_norm(x)

        return x, attn, pos_bias



#! /usr/bin/python
# -*- encoding: utf-8 -*-
# Adapted from https://github.com/wujiyang/Face_Pytorch (Apache License)

import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from .utils import accuracy


class LossFunction(nn.Module):
    def __init__(self, nOut, nClasses, margin=0.3, scale=15, easy_margin=False, **kwargs):
        super(LossFunction, self).__init__()

        self.test_normalize = True
        
        self.m = margin
        self.s = scale
        self.in_feats = nOut
        self.weight = torch.nn.Parameter(torch.FloatTensor(nClasses, nOut), requires_grad=True)
        self.ce = nn.CrossEntropyLoss()
        nn.init.xavier_normal_(self.weight, gain=1)

        self.easy_margin = easy_margin
        self.cos_m = math.cos(self.m)
        self.sin_m = math.sin(self.m)

        # make the function cos(theta+m) monotonic decreasing while theta in [0°,180°]
        self.th = math.cos(math.pi - self.m)
        self.mm = math.sin(math.pi - self.m) * self.m

        # print('Initialised AAMSoftmax margin %.3f scale %.3f'%(self.m,self.s))

    def forward(self, x, label=None):

        assert x.size()[0] == label.size()[0]
        assert x.size()[1] == self.in_feats
        
        # cos(theta)
        cosine = F.linear(F.normalize(x), F.normalize(self.weight))
        # cos(theta + m)
        sine = torch.sqrt((1.0 - torch.mul(cosine, cosine)).clamp(0, 1))
        phi = cosine * self.cos_m - sine * self.sin_m

        if self.easy_margin:
            phi = torch.where(cosine > 0, phi, cosine)
        else:
            phi = torch.where((cosine - self.th) > 0, phi, cosine - self.mm)

        # one_hot = torch.zeros(cosine.size(), device='cuda' if torch.cuda.is_available() else 'cpu')
        one_hot = torch.zeros_like(cosine)
        one_hot.scatter_(1, label.view(-1, 1), 1)
        output = (one_hot * phi) + ((1.0 - one_hot) * cosine)
        output = output * self.s

        loss    = self.ce(output, label)
        prec1   = accuracy(output.detach(), label.detach(), topk=(1,))[0]
        return loss, prec1


from .DataLoader import DataLoader
import torch
import os
from huggingface_hub import hf_hub_download
from .model_args import get_default_param
from .SpeakerNet import SpeakerNet

os.environ['CUDA_VISIBLE_DEVICES'] = '0'

args = get_default_param()

model_path = '/home/$USER/Downloads/models'
cache_dir = '/home/$USER/.cache/huggingface/hub/models--thejan-fonseka--DeepSpeakerVerifier'

model_head_mapping = {
    4: "model_h4_dim.model",
    8: "model_h8_dim.model",
    16: "model_h16_dim.model"
}

model_name_download_mapping = {
    "model_h4_dim.model": "https://huggingface.co/thejan-fonseka/DeepSpeakerVerifier/blob/main/model_h4_dim.model",
    "model_h8_dim.model": "https://huggingface.co/thejan-fonseka/DeepSpeakerVerifier/blob/main/model_h8_dim.model",
    "model_h16_dim.model": "https://huggingface.co/thejan-fonseka/DeepSpeakerVerifier/blob/main/model_h16_dim.model",
}

model_threshold = {
    4: 0.4,
    8: 0.39,
    16: 0.37
}


class ModelLoader:
    """
    ModelLoader class for loading and working with deep learning models.

    get_embedding:
    Get the embedding for the audio data provided.

    Args:
        audio_path: Path to the audio file.
        evamode: Flag indicating evaluation mode (default is True).
        num_eval: Number of evaluations to perform (default is 10).
        vad: Flag indicating whether to use VAD mode (default is True).

    Returns:
        Embedding of the audio data.

    load_model:
    Load a model from a specified path or download it if not available locally.

    Args:
        model: The model object to load.
        model_name: Name of the model.
        model_path: Path to the directory where the model is stored.
        cache_dir: Directory for caching (default is cache_dir).

    Returns:
        None.

    get_threshold:
    Get the threshold for the model.

    loadParameters:
    Load parameters into the model.
    """

    def __init__(self, model_name, attention_heads=8):
        args["attention_heads"] = attention_heads
        self.model_name = model_head_mapping[attention_heads]
        self.model_path = model_path
        self.model =SpeakerNet(**args).to(args['device'])
        self.load_model(self.model, self.model_name, self.model_path)

    def get_embedding(self, audio_path, evamode=True, num_eval=10, vad=True):
        """
        Get the embedding for the audio data provided.

        Args:
            audio_path: Path to the audio file.
            evamode: Flag indicating evaluation mode (default is True).
            num_eval: Number of evaluations to perform (default is 10).
            vad: Flag indicating whether to use VAD mode (default is True).

        Returns:
            Embedding of the audio data.
        """
        feats = DataLoader.load_audio(filename=audio_path, evalmode=evamode, num_eval=num_eval, vad_mode=vad)
        if feats.size == 0:
            return None
        feats = torch.FloatTensor(feats)
        embedding = self.model([feats, 'test'])
        embedding = embedding.detach().cpu().numpy()
        return embedding

    def load_model(self, model, model_name, model_path, cache_dir=cache_dir):

        """
        Load a model from a specified path or download it if not available locally.

        Args:
            model: The model object to load.
            model_name: Name of the model.
            model_path: Path to the directory where the model is stored.
            cache_dir: Directory for caching (default is cache_dir).

        Returns:
            None.
        """
        model_file = os.path.join(model_path, model_name)

        if not os.path.exists(model_file):
            # Download the model from Hugging Face if not available locally
            model_file = hf_hub_download(repo_id="thejan-fonseka/DeepSpeakerVerifier", filename=model_name)
            
        # Load the model from the local file
        model.load_state_dict(torch.load(model_file, map_location=torch.device(args['device'])))        
        # self.loadParameters(torch.load(model_file, map_location=torch.device(args['device'])))
        model.eval()
    
    def get_threshold(self):
        """
        Get the threshold for the model.
        """
        return model_threshold[args["attention_heads"]]
    
    def loadParameters(self, param):
        """
        Load parameters into the model.
        """
        self_state = self.model.state_dict();
        loaded_state = param

        for name, param in loaded_state.items():
            origname = name;
            
            if name not in self_state:
                # print("%s is not in the model."%origname);
                continue;
            if self_state[name].size() != loaded_state[origname].size():
                print("Wrong parameter length: %s, model: %s, loaded: %s"%(origname, self_state[name].size(), loaded_state[origname].size()));
                continue;
            self_state[name].copy_(param);


from sklearn.metrics.pairwise import cosine_similarity

class Similarity:
    """
    Similarity class for calculating similarity metrics.

    cosine_similarity:
    Calculate the cosine similarity between two vectors.

    Args:
        a: First vector.
        b: Second vector.

    Returns:
        Mean of the cosine similarity between the two vectors.
    """

    
    @staticmethod
    def cosine_similarity(a, b):
        """
        Calculate the cosine similarity between two vectors.

        Args:
            a: First vector.
            b: Second vector.

        Returns:
            Mean of the cosine similarity between the two vectors.
        """
    
        return cosine_similarity(a, b).mean()



import random
import webrtcvad
import soundfile as sf
import numpy as np
import librosa
import struct


class DataLoader:
    """
    Load audio data from a file and process it for further analysis.

    Args:
        filename: Path to the audio file.
        max_frames: Maximum number of frames to consider (default is 300).
        evalmode: Flag indicating evaluation mode (default is True).
        num_eval: Number of evaluations to perform (default is 10).
        vad_mode: Flag indicating whether to use VAD mode (default is True).

    Returns:
        Processed audio data ready for analysis.
    """

    @staticmethod
    def load_audio(filename: str, max_frames: int = 300, evalmode=True, num_eval=10, vad_mode=True):
        """
        Load audio data from a file and process it for further analysis.

        Args:
            filename: Path to the audio file.
            max_frames: Maximum number of frames to consider (default is 300).
            evalmode: Flag indicating evaluation mode (default is True).
            num_eval: Number of evaluations to perform (default is 10).
            vad_mode: Flag indicating whether to use VAD mode (default is True).

        Returns:
            Processed audio data ready for analysis.
        """
        max_audio = max_frames * 160 + 240
        
        audio, sample_rate = sf.read(filename)
        # audio = nr.reduce_noise(y=audio, sr=sample_rate)

        if vad_mode:
            # Create a VAD instance
            vad = webrtcvad.Vad()
            vad.set_mode(3)  # Set the aggressiveness mode (0-3)

            # Initialize an array to store the speech frames
            speech_frames = []

            # Process the audio in 30ms frames with a 10ms step
            frame_duration = 0.03  # 30ms
            frame_stride = 0.01  # 10ms
            frames = librosa.util.frame(audio, frame_length=int(frame_duration * sample_rate),
                                        hop_length=int(frame_stride * sample_rate))
            
            for frame in frames.T:
                frame_bytes = struct.pack(f"{len(frame)}h", *np.rint(frame * 32767).astype(np.int16))
                if vad.is_speech(frame_bytes, sample_rate):
                    speech_frames.append(frame)

            # Concatenate the speech frames into a single array
            if speech_frames:
                speech_data = np.concatenate(speech_frames)
            else:
                speech_data = np.array([])
        # else:
        speech_data = audio

        # Handle empty speech data
        if speech_data.size == 0:
            return speech_data

            # if evalmode:
            #     return np.zeros((num_eval, max_audio), dtype=np.float64)
            # else:
            #     return np.zeros((num_eval, max_audio), dtype=np.float64)

        audiosize = speech_data.shape[0]

        if audiosize <= max_audio:
            shortage = max_audio - audiosize + 1
            speech_data = np.pad(speech_data, (0, shortage), 'wrap')
            audiosize = speech_data.shape[0]

        if evalmode:
            startframe = np.linspace(0, audiosize - max_audio, num=num_eval)
        else:
            startframe = np.array([np.int64(random.random() * (audiosize - max_audio))])

        feats = []
        if evalmode and max_frames == 0:
            feats.append(speech_data)
        else:
            for asf in startframe:
                feats.append(speech_data[int(asf):int(asf) + max_audio])

        feat = np.stack(feats, axis=0).astype(np.float64)
        return feat

if __name__ == "__main__":
    f = DataLoader.load_audio("../samples/dr-uthaya-e1.mp3", 160)
    print(f)
